#ifndef CLUTTER_CANCELLATION_H
#define CLUTTER_CANCELLATION_H

#include <algorithm>
#include <cassert>
#include <complex>
#include <iostream>
#include <stdexcept>
#include <valarray>
#include <vector>

#include "fft.h"
#include "numpy_helper.h"

using std::cerr;
using std::complex;
using std::copy;
using std::invalid_argument;
using std::valarray;
using std::vector;

typedef vector<complex<double>> array;
typedef vector<array> mat;

/*
 * Description:
 * Generate a signal subspace parameter list for the ECA algorithm variants.
 * 
 * Parameters:
 * K: Number of considered time delay
 * D: Maximum Doppler frequency value 
 */
vector<vector<int>> gen_subspace_indexes(int K, int D);

/*
 * Description:
 * Performs filtering on the surveillance channel in the time domain. The ECA-S algorithm is a modification
 * of the ECA-B alg. where the filtering and the coefficient estimation is made on different block sizes.
 * The window used for the coefficient estimation is the symmetrical extension of the window that is used for the
 * filtering. This extension can be controlled via the Na parameter. Na is the one sided window extension
 * measured in signal samples.
 * 
 * Parameters:
 * ref_ch : Signal component to suppress (reference channel)
 * surv_ch : Channel to filter (surveillance channel)
 * subspace_list : Contains the time delay and Doppler shift which is used to span the signal subspace
 * T : Number of batches
 * Na : One sided window extension measured in samples
 */
array ECAS(array& ref_ch, array& surv_ch, vector<vector<int>>& subspace_list, int T, int Na);


/*
 * Description:
 * Calculates the part of the correlation function of arrays with same size
 * The total length of the cross-correlation function is 2*N-1, but this
 * function calculates the values of the cross-correlation between [N-1 : N+clen-1]
 * 
 * Parameters:
 * y : input array
 * x : input array
 * clen : correlation length
 */
array pruned_correlation(array& y, array& x, int clen);


/*
 * Description:
 * Performs Wiener filtering with applying the Minimum Redundance Estimation (MRE) technique. 
 * When using MRE, the autocorrelation matrix is not fully estimated, but only the first column.
 * With this modification the required calculations can be reduced from KxK to K element.
 * 
 * Parameters:
 * ref_ch : Reference signal array
 * surv_ch : Surveillance signal array
 * K : Filter tap number
 */
array Wiener_SMI_MRE(array& ref_ch, array& surv_ch, int K);   

#endif