#ifndef NUMPY_HELPER_H
#define NUMPY_HELPER_H

#include <cassert>
#include <complex>
#include <iostream>
#include <valarray>
#include <vector>
#include <Eigen/Eigenvalues> 

using std::cerr;
using std::complex;
using std::conj;
using std::valarray;
using std::vector;


typedef vector<complex<double>> array;
typedef vector<array> mat;

inline void arange(vector<int>& arr, int start, int stop, int step) {
    if (arr.size()) {
        arr.clear();
    }

    int new_size = (stop - start) / step;
    if ((stop - start) % step) {
        ++new_size;
    }

    arr.resize(new_size);
    int idx = 0;
    for (int i = start; i < stop; i += step) {
        arr[idx++] = i; 
    }

    assert(idx == int(arr.size()));
}

template<class T>
T dot(const vector<T>& vec1, const vector<T>& vec2) {
    assert(vec1.size() == vec2.size());

    T sum = 0;
    for (int i = 0; i < static_cast<int>(vec1.size()); ++i) {
        sum += vec1[i] * vec2[i];
    }

    return sum;
}

inline array dot(const mat& matrix, const array& arr) {
    assert(matrix[0].size() == arr.size());

    array output(matrix.size());
    for (int i = 0; i < static_cast<int>(matrix.size()); ++i) {
        output[i] = dot(matrix[i], arr);
    }

    return output;
}

inline array conjugate(const array& vec) {
    array output(vec.size());

    for (int i = 0; i < static_cast<int>(vec.size()); ++i) {
        output[i] = conj(vec[i]);
    }
    
    return output;
}

inline complex<double> sum(const mat& matrix) {
    complex<double> sum(0.0, 0.0);
    
    for (auto& row : matrix) {
        for (auto& col : row) {
            sum += col;
        }
    }

    return sum;
}

inline mat conjugate(const mat& matrix) {
    int rows = static_cast<int>(matrix.size());
    int cols = static_cast<int>(matrix[0].size());
    mat output(rows, array(cols));

    for (int i = 0; i < rows; ++i) {
        for (int j = 0; j < cols; ++j) {
            output[i][j] = conj(matrix[i][j]);
        }
    }

    return output;
}

inline mat reshape(const array& vec, int num_rows, int num_cols) {
    assert(num_rows * num_cols == static_cast<int>(vec.size()));

    mat output(num_rows, array(num_cols));
    int k = 0;
    for (int i = 0; i < num_rows; ++i) {
        for (int j = 0; j < num_cols; ++j) {
            output[i][j] = vec[k++];
        }
    }

    return output;
}

// Similar to np.roll, but does not circularly shift values
inline array shift(const array& x, int i) {
    int N = static_cast<int>(x.size());
    if (!i) {
        return x;
    }

    if (i > N) {
        return array(N, complex<double>(0, 0));
    }

    array output(N, complex<double>(0, 0));
    for (int idx = N - 1; idx >= i; --idx) {
        output[idx] = x[idx - i];
    }

    return output;
}

inline mat transpose(const mat& x) {
    mat outputs(x[0].size(), array(x.size()));
    int N = static_cast<int>(x.size());
    int M = static_cast<int>(x[0].size());
    for (int i = 0; i < N; ++i) {
        for (int j = 0; j < M; ++j) {
            outputs[j][i] = x[i][j];
        }
    }

    return outputs;
}

inline mat transpose(mat& x) {
    mat outputs(x[0].size(), array(x.size()));
    int N = static_cast<int>(x.size());
    int M = static_cast<int>(x[0].size());
    for (int i = 0; i < N; ++i) {
        for (int j = 0; j < M; ++j) {
            outputs[j][i] = x[i][j];
        }
    }

    return outputs;
}

inline mat transposeConj(const mat& x) {
    mat outputs(x[0].size(), array(x.size()));
    int N = static_cast<int>(x.size());
    int M = static_cast<int>(x[0].size());
    for (int i = 0; i < N; ++i) {
        for (int j = 0; j < M; ++j) {
            outputs[j][i] = conj(x[i][j]);
        }
    }

    return outputs;
}

inline mat matMult(const mat& lhs, const mat& rhs) {
    int resultRows = static_cast<int>(lhs.size());
    int resultCols = static_cast<int>(rhs[0].size());
    assert(lhs[0].size() == rhs.size());
    int middle = static_cast<int>(rhs.size());
    mat output(resultRows, array(resultCols));

    for (int i = 0; i < resultRows; ++i) {
        for (int j = 0; j < resultCols; ++j) {
            for (int k = 0; k < middle; ++k) {
                output[i][j] += lhs[i][k] * rhs[k][j]; 
            }
        }
    }

    return output;
}

inline mat multiply(const mat& lhs, const mat& rhs) {
    int rows = static_cast<int>(lhs.size());
    int cols = static_cast<int>(lhs[0].size());

    assert(static_cast<int>(rhs.size()) == rows && static_cast<int>(rhs[0].size()) == cols);
    mat output(rows, array(cols));

    for (int i = 0; i < rows; ++i) {
        for (int j = 0; j < cols; ++j) {
            output[i][j] = lhs[i][j] * rhs[i][j];
        }
    }

    return output;
}

inline mat add(const mat& x, const mat& y) {
    int rows = static_cast<int>(x.size());
    int cols = static_cast<int>(y.size());

    assert(x.size() == y.size() && x[0].size() == y[0].size());

    mat output(rows, array(cols));
    for (int i = 0; i < rows; ++i) {
        for (int j = 0; j < cols; ++j) {
            output[i][j] = x[i][j] + y[i][j];
        }
    }

    return output;
}

inline array convolve(const array& x, const array& y) {
    int N1 = static_cast<int>(x.size());
    int N2 = static_cast<int>(y.size());

    array output(N1 + N2 - 1);
    for (int i = 0; i < N1 + N2 - 1; i++){
        output[i] = 0;
        for (int j = 0; j < N2; j++){
            if((i - j >= 0) && (i - j < N1)){ //make sure we're not indexing out of bounds
                output[i] += x[i - j] * y[j];
            }
        }
    }

    return output;
}


inline mat getCofactor(const mat& x, int p, int q) {
    assert(x.size() == x[0].size());
    int n = static_cast<int>(x.size());

    mat output(n - 1, array(n - 1));

    int row = 0, col = 0;
    for (int i = 0; i < n; ++i) {
        for (int j = 0; j < n; ++j) {
            if (i != p && j != q) {
                output[row][col++] = x[i][j];

                if (col == n - 1) {
                    col = 0;
                    row++;
                }
            }
        }
    }

    return output;
}


inline complex<double> determinant(const mat& x) {
    assert(x.size() > 0 && x.size() == x[0].size());
    int n = static_cast<int>(x.size());

    Eigen::MatrixXcd temp(n, n);
    for (int i = 0; i < n; i++){
        for (int j = 0; j < n; j++){
             temp.row(i)[j] = x[i][j];
        }
    }

    return temp.determinant();
}


inline mat adjoint(const mat& x) {
    int n = static_cast<int>(x.size());
    assert(n > 1 && x.size() == x[0].size());
    mat output(n, array(n));

    double sign = 1;

    for (int i = 0; i < n; ++i) {
        for (int j = 0; j < n; ++j) {
            sign = ((i + j) % 2 == 0) ? 1 : -1;
            output[j][i] = sign * (determinant(getCofactor(x, i, j)));
        }
    }

    return output;
}

inline mat inverse(const mat& x) {
    int n = static_cast<int>(x.size());
    assert(n > 1 && x.size() == x[0].size());

    Eigen::MatrixXcd temp(n, n);
    for (int i = 0; i < n; i++){
        for (int j = 0; j < n; j++){
             temp.row(i)[j] = x[i][j];
        }
    }

    temp = temp.inverse();

    mat output(n, vector<complex<double> >(n));
    for (int i = 0; i < n; i++){
        for (int j = 0; j < n; j++){
             output[i][j] = temp.row(i)[j];
        }
    }

    return output;
}

inline mat fftshift(const mat& x) {
    int rows = static_cast<int>(x.size());
    int cols = static_cast<int>(x[0].size());
    mat output = x;
    int row_shift = (rows + 1) / 2;
    int column_shift = (cols + 1) /2;
    for (int i = 0; i < rows; i++) {
        for (int j = 0; j < cols; j++) {
            output[i][j] = x[(i+row_shift) % rows][(j+column_shift) % cols];
        }
    }

    return output;
}

#endif