#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from iq_header import IQHeader
import numpy as np
import glob
import logging
import os
from matplotlib import pyplot as plt

import pyapril.channelPreparation as chprep
from pyapril import detector
import pyapril.metricExtract as me
from  pyapril import hitProcessor
from pyapril.RDTools import export_rd_matrix_img
"""
    This script performs reference signal processing on a standard VEGA dataset compatible measurement.
    During the reference processing, the script automatically loads the recorded IQ frames ony by one and performs
    implemented passive radar DSP tasks...TBD
    
    To use the processing script a prepared target reference track array is required. The format of this
    array is the following: (*.tpt)

    +------------+-----------+----------+-----------+----------+-------+-----------+-------+---------+---------+
    | Time index | Timestamp | Latitude | Longitude | Altitude | Speed | Direction | Range | Doppler | Azimuth |
    +------------+-----------+----------+-----------+----------+-------+-----------+-------+---------+---------+
    |            |           |          |           |          |       |           |       |         |         |
    +------------+-----------+----------+-----------+----------+-------+-----------+-------+---------+---------+
   
    This array can be automatically generated from FlightRadar24 tracks for a VEGA dataset compatible measurement
    using the 'FR24_track_preroc' script, you can find it in the "VEGA_databse_tools" repository on github.
    
    The saved reference processing result file has the following format (*.trt)
    +------------+-------+-------+---------+----------+
    | Time index |  SINR |  Range | Doppler | Azimuth |
    +------------+-------+--------+---------+---------+
    |            |       |        |         |         |
    +------------+-------+-------+---------+----------+


"""

def load_iq(file_name):
    """
        Description: 
        ------------
        Load IQ frame and prepare the payload section for further processings
        
        Parameters:
        -----------
        :param: file_name: Filename which stores the recorded IQ frame with 
                          ".iqf" extension.
        :type: file_name : string
        
        Return values:
        --------------
        :return: iq_samples: IQ sample matrix extracted from the IQ frame
        :return: iq_header : IQ header extracted from the IQ frame
        
        :rtype: iq_samples: M x N complex numpy array 
        :rtype: iq_header : IQ header object 
            
    """
    file_descr = open(file_name, "rb")

    iq_header_bytes = file_descr.read(1024)
    iq_header = IQHeader()
    iq_header.decode_header(iq_header_bytes)

    iq_data_length = int((iq_header.cpi_length * iq_header.active_ant_chs * (2*iq_header.sample_bit_depth))/8)
    iq_data_bytes = file_descr.read(iq_data_length)

    file_descr.close()
          
    iq_cf64 = np.frombuffer(iq_data_bytes, dtype=np.complex64).reshape(iq_header.active_ant_chs, iq_header.cpi_length)
    iq_samples = iq_cf64.copy()
    
    return iq_samples, iq_header

def find_target(rd_matrix, range_min, range_max, Doppler_min, Doppler_max):
    """
        Description:
        ------------
        This function tries to find the position of target reflection peak on the given range-Doppler matrix
        inside a specified range-Doppler region. To find the true target location, the function search for
        the maximum value inside this region, which is bounded by the 
        'range_min'-'range_max' and 'Doppler_min'-'Doppler_max' matrix indicies.
        
        Parameters:
        -----------
        :param: rd_matrix  : range-Doppler matrix
        :param: range_min  : minimum range cell index
        :param: range_max  : maximum range cell index
        :param: Doppler_min: minimum Doppler cell index
        :param: Doppler_max: maximum Doppler cell index
        
        :type: rd_matrix  : R x D complex numpy array
        :type: range_min  : int
        :type: range_max  : int
        :type: Doppler_min: int
        :type: Doppler_max: int  
            
        Return values:
        --------------
        :return: target_range_cell  :Range cell of the found target refleciton
        :return: target_doppler_cell:Doppler cell of the found target reflection
        
        :rtype: target_range_cell  : int
        :rtype: target_doppler_cell: int
    """
            
    rd_matrix = np.abs(rd_matrix)
    
    # Range and Doppler coordinates
    (Doppler_cor, range_cor) = np.unravel_index(np.argmax(rd_matrix[Doppler_min:Doppler_max, range_min:range_max]),
                                              rd_matrix[Doppler_min:Doppler_max, range_min:range_max].shape)

    target_range_cell       = range_cor + range_min
    target_doppler_cell     = Doppler_cor + Doppler_min
    
    return (target_range_cell, target_doppler_cell)

def print_target_reference_track(target_reference_track):
    fi = open("target_reference_track.txt", 'w')
    fi.write('{} {}\n'.format(target_reference_track.shape[0], target_reference_track.shape[1]))
    for i in range(target_reference_track.shape[0]):
        for j in range(target_reference_track.shape[1]):
            fi.write("{} ".format(target_reference_track[i][j]))
        fi.write("\n")

def print_iq_samples(iq_samples, fs, t):
    fi = open('dataset_{}.txt'.format(t), 'w')

    fi.write('{}\n'.format(fs))
    fi.write('{} {}\n'.format(iq_samples.shape[0], iq_samples.shape[1]))

    for i in range(iq_samples.shape[0]):
        for j in range(iq_samples.shape[1]):
            fi.write('{} {} '.format(iq_samples[i][j].real, iq_samples[i][j].imag))

        fi.write('\n')

"""
---------------------------
                           
    P A R A M E T E R S    
                            
---------------------------
"""
# -> Processing parameters
ref_ch_index = 0
td_filter_method= "SMI-MRE"
K=128 # Time domain filter dimension
cc_det_window = 'Hann' # Windowi function used for the Range-Doppler processing
DOA_method = "MUSIC"
M = 3 # Number of antenna channels
d = 0.528 # Antenna inter-element spacing

# -> Detection performance metric extraction parameters
win=[6,6,3,3]
win_pos=[30,50]
threshold = 10
search_window_size = 8 # One sided

# -> Set measurement base paths
logging.basicConfig(level=logging.INFO)
vega_measurement_path= "/home/bingscha/VEGAdatabasetools"
iq_folder_path       = os.path.join(vega_measurement_path,"iq")
target_info_path     = os.path.join(vega_measurement_path,"target_info")
results_path         = os.path.join(vega_measurement_path,"results")
track_fname_temp     = 'target_ref_proc_track_'

iqf_files = glob.glob(os.path.join(iq_folder_path,"*.iqf"))
fname_prefix = iqf_files[0].split('_')[0]+"_"
import pdb; pdb.set_trace()
trt_files = glob.glob(os.path.join(target_info_path,"*.trt"))

target_index = 0
logging.warning("Processing only one target track, ID: {:d}".format(target_index))
target_referece_track = np.loadtxt(trt_files[target_index])
print_target_reference_track(target_referece_track)
# -> Set additional processing parameters
iq_samples, iq_header = load_iq(iqf_files[0])

start_file_index = int(target_referece_track[0,0])
stop_file_index  = int(target_referece_track[-1,0])

max_range   = np.max(target_referece_track[:,7])
max_Doppler = np.max(target_referece_track[:,8])

# Get sampling frequency
fs = iq_header.sampling_freq

# Get CPI size
N = np.size(iq_samples, 1)

# Doppler resolution
fD_res = fs / (2*N)

range_res = 3*10**8/fs
Doppler_cells = int(max_Doppler/fD_res)+search_window_size+win[0]+1
max_Doppler_ext = np.ceil(Doppler_cells*fD_res)
max_range_ext = int(2**(np.ceil(np.log2((max_range/range_res)+search_window_size+win[0]))))
logging.info("range-Doppler matrix size will be: {:d} bins x {:f} Hz ".format(max_range_ext, max_Doppler_ext))

# Antenna array alignment vector
array_alignment = np.arange(M)*d 

# Allocate Processed Range-Doppler-Azimuth track array
target_processed_track = np.zeros((np.size(target_referece_track,0),5), dtype=float)

rd_maps_for_doa = None # Allocated with the first processing
for t in np.arange(0,stop_file_index-start_file_index+1,1):  
    """
    ---------------------
           L O A D
    ---------------------    
    """
    file_name = (fname_prefix+str(start_file_index+t)+".iqf")
    iq_samples, iq_header = load_iq(file_name)
    ref_target_range_cell   = int(np.round(target_referece_track[t, 7]/range_res))
    ref_target_Doppler_cell = int(np.round(target_referece_track[t, 8]/fD_res)+Doppler_cells)
    ref_target_doa_dir      = target_referece_track[t, 9]
    # Uniform linear array assumption
    if ref_target_doa_dir < 0:
        ref_target_doa_dir += 180
    if ref_target_doa_dir > 180:
        ref_target_doa_dir -= 180
    """
    ---------------------
     P R O C E S S I N G
    ---------------------    
    """
    logging.info("Processing: {:s}".format(file_name.split('/')[-1]))
    
    # Separate reference and surveillance channels
    ref_ch, surv_chs = chprep.isolate_channels(iq_samples, ref_ch_index)
    print_iq_samples(iq_samples, fs, start_file_index+t)
    
    # Beamforing
    surv_ch ,w = chprep.beamform_surveillance(surv_chs.T, array_alignment, "MVDR",
                            direction = ref_target_doa_dir)
    
    # Filtering surveillance channel
    filt_surv_ch = chprep.time_domain_filter_surveillance(ref_ch, surv_ch, td_filter_method, K=K)
    
    # Windowing
    windowed_surv_ch = detector.windowing(filt_surv_ch, cc_det_window)
    
    # Apply Cross-correlation detector    
    rd_matrix = detector.cc_detector_ons(ref_ch, windowed_surv_ch, fs, max_Doppler_ext,
                                        int(max_range_ext), verbose=0, Qt_obj=None)
    
    # Try to find target reflection peak
    target_range_cell, target_doppler_cell =  find_target(rd_matrix,
                                                          ref_target_range_cell-search_window_size,
                                                          ref_target_range_cell+search_window_size,
                                                          ref_target_Doppler_cell-search_window_size,
                                                          ref_target_Doppler_cell+search_window_size)
    
    
    # Target Azimuth estimation
    hit_list=[]
    hit_list.append([target_range_cell, target_doppler_cell])
    if t == 0:
        M=np.size(surv_chs,0)   # Number of surveillance channels
        D=np.size(rd_matrix,0) # Number of Doppler cells
        R=np.size(rd_matrix,1) # Number of range cells 
        rd_maps_for_doa = np.zeros([M, D, R], dtype=complex)
    
    for m in range(M):
         # Filtering surveillance channel
        filt_surv_ch = chprep.time_domain_filter_surveillance(ref_ch, surv_chs[m ,:], td_filter_method, K=K)
        
        # Windowing
        windowed_surv_ch = detector.windowing(filt_surv_ch, cc_det_window)
        
        # Apply Cross-correlation detector    
        rd_maps_for_doa[m,:,:] = detector.cc_detector_ons(ref_ch, windowed_surv_ch, fs, max_Doppler_ext,
                                            int(max_range_ext), verbose=0, Qt_obj=None)
    estimated_doas = hitProcessor.target_DOA_estimation(rd_maps_for_doa, 
                                                        hit_list,
                                                        DOA_method, 
                                                        array_alignment)                                                        
       
    logging.info("Target range cell ref/found [{:d}/{:d}]".format(ref_target_range_cell,target_range_cell))
    logging.info("Target Doppler cell ref/found [{:d}/{:d}]".format(ref_target_Doppler_cell,target_doppler_cell))
    logging.info("Target azimuthl ref/found [{:.1f}/{:.1f}]".format(target_referece_track[t, 9],estimated_doas[0]))
    
    # Extracting SNR and save results
    target_rd=[target_range_cell, target_doppler_cell]
    
    target_rd[1] -= (np.size(rd_matrix,0)-1)//2
    win_pos[1] -= ( np.size(rd_matrix,0)-1)//2
    
    delta = me.extract_delta(rd_matrix, target_rd, win)
    logging.info("Target SINR {:.1f}".format(delta))    

    if delta < threshold:
        box_range_index = ref_target_range_cell
        box_Doppler_index = ref_target_Doppler_cell
        box_color = 'red'
    else:
        box_range_index = target_range_cell
        box_Doppler_index = target_doppler_cell
        box_color = 'lime'

    
    rd_png_fname= "RD_matrix_{:04d}.png".format(t)   
    
    export_rd_matrix_img(rd_png_fname, rd_matrix, max_Doppler_ext,
                         box_range_index,
                         box_Doppler_index,
                         box_color,
                         search_window_size)
    
    # Store RDA track
    target_processed_track[t, 0] = start_file_index+t
    target_processed_track[t, 1] = delta
    target_processed_track[t, 2] = target_range_cell*range_res
    target_processed_track[t, 3] = (target_doppler_cell-Doppler_cells)*fD_res
    target_processed_track[t, 4] = estimated_doas[0]
    
        
logging.info("Saving processed target track array for target ID: {:d}".format(target_index))
fname = os.path.join(results_path, track_fname_temp+str(target_index)+".tpt")
np.savetxt(fname, target_processed_track)




