#include "beamform.h"
#include "numpy_helper.h"

#include <vector>
#include <complex>       
#include <math.h>
#include <algorithm>
#include <iostream>

using namespace std;

typedef vector<vector<complex<double>>> mat;


mat estimate_corr_matrix(mat& X) {
    complex<double> N(X.size(), 0);
    int M = X[0].size();
    mat R(M, vector<complex<double>>(M,0));

    if (N.real() < M) {
        cout << "WARNING: Number of antenna elements is greather than the number of time samples" << endl;
        cout << "WARNING: You may flipped the input matrix" << endl;
        exit(1);
    }
    
    mat t = transpose(X);
    R = matMult(t, transpose(conjugate(t)));
    for (int i = 0; i<M; i++) {
        for (int j = 0; j<M; j++) {
            R[i][j] = R[i][j]/N;
        }
    }   
    return R;
}

/*                                 
                Optimal soluation of the Wiener filtering for beamforiming

    Description:
    ------------    
        This function calculates the optimum Wiener solution. To obtain this it uses the
        array resonse of the signal of intereset and the autocorrelation matrix of the interference
        plus noise signals. In many real situation this is substituted with the autocorrelation matrix 
        of the received signal which is comprised of the signal of inteters, the interferences and the noise.
        "Frank B Gross: Smart Antennas With MATLAB, 2nd edition 2015 , section 8.3.1"   
        
        Note that the optimal Wiener beamformer coincides with MSINR or the MVDR beamformer coefficents
        excpeting a scalar multiplier value!
        
        To calculate the coefficients the incident angle of the desired signal must be known 
        preliminary.
        
    Parameters:
    -----------           
        Rnunu : (M x M complex numpy matrix) Autocorrelation matrix of the noise + interference signals                    
        aS    : (M x 1 complex numpy vector) Array response vector of the signal of interest 
   
   Return values:
    -----------------            
        
        w_msinr: (complex numpy array) Calculated coefficient vector    
        None   : Input data was not consistens
*/
mat optimal_Wiener_beamform(mat& Rnunu, mat& aS) {
    if (Rnunu.size() != aS.size()) {
        cout << "ERROR: Input data matrices do not have consistent shape" << endl;
        exit(1);
    }

    if (Rnunu.size() == 0 || Rnunu.size() != Rnunu[0].size()) {
        cout << "ERROR: The correlation matrix is not squar" << endl; 
        exit(1);
    }

    if (aS.size() == 0 || aS[0].size() != 1) {
        cout<< "ERROR: aS must be a vector" << endl;
        exit(1);
    }

    auto inv = inverse(Rnunu);

    return matMult(inv, aS);
}


// Description:
// -----------
//     The underlying function produces the signal of the surveillance channel 
//     from the antenna channels with the application of beamspace processing
//     techniques.
vector<complex<double>> beamform_surveillance(mat& ant_chs, vector<double>& alignment_vector, double direction) {
    int N = ant_chs.size();
    int M = ant_chs[0].size();
    int align_size = alignment_vector.size();
    
    if (M > N) {
        cout << "WARNING: The number of antenna channels is greater than the number of samples in a channel. M > N" << endl;
        exit(1);
    }
    if (M != align_size) {
        cout <<"ERROR: Mismatch in the input signal channel numbers and the antenna alignment vector size"<<endl;
        exit(1);
    }

    mat R = estimate_corr_matrix(ant_chs);
    const std::complex<double> i(0, 2*M_PI);
    mat aS(alignment_vector.size(),vector<complex<double>>(1));
    for (int j = 0; j < align_size; j++) {
        aS[j][0] = exp(alignment_vector[j] *  i * cos(0.0174533*direction));
    }
    
    mat w = optimal_Wiener_beamform(R, aS);
    int w_row = w.size();
    int w_col = w[0].size();
    double sum = 0;
    for (int j = 0; j < w_row; j++) {
        for (int k = 0; k < w_col; k++) {
            sum += abs(w[j][k]);
        }
    }
    double norm_factor = sqrt(M) / sum;
    for (int j = 0; j < w_row; j++) {
        for (int k = 0; k < w_col; k++) {
            w[j][k] *= norm_factor;
        }
    }   

    mat surv_ch = matMult( transpose(conjugate(w)), transpose(ant_chs) );

    return surv_ch[0];
}

