#include "clutterCancellation.h"

using std::vector;

vector<vector<int>> gen_subspace_indexes(int K, int D) {
    vector<vector<int>> output((D + 1) * K, vector<int>(2));

    int count = 0;
    for (int i = -D; i < D + 1; ++i) {
        for (int j = 0; j < K; ++j) {
            output[count][0] = i;
            output[count++][1] = j;
        }
    }

    return output;
}

array ECAS(array& ref_ch, array& surv_ch, vector<vector<int>>& subspace_list, int T, int Na) {

    int N = static_cast<int>(ref_ch.size());
    int batch_size = N / T;

    // Input checks
    if (N % T) {
        throw invalid_argument("ERROR: Improper batch size, filtering is bypassed\nERROR: No output is generated\n");
    }

    for (int i = 0; i < static_cast<int>(subspace_list.size()); ++i) {
        if (subspace_list[i][1] > batch_size) {
            throw invalid_argument("ERROR: Time delay is greater than the size of the batch, \
                                    this is not handled properly. Time delay is bypassed\n   \
                                    ERROR: No output is generated\n");
        }
        if (subspace_list[i][1] > batch_size + Na) {
            throw invalid_argument("ERROR: Time delay is greater than the size of the batch plus the sliding window, \
                                    this is not handled properly. Time delay is bypassed\n   \
                                    ERROR: No output is generated\n");
        }
    }

    // Calculations
    int KD = static_cast<int>(subspace_list.size());

    array extended_ref_ch(N + 2 * Na);
    array extended_surv_ch(N + 2 * Na);

    copy(extended_ref_ch.begin() + Na, extended_ref_ch.begin() + (Na + N), ref_ch.begin());
    copy(extended_surv_ch.begin() + Na, extended_surv_ch.begin() + (Na + N), surv_ch.begin());

    array filtered_surv_ch(N);
    for (int l = 0; l < N; l += batch_size) {
        array surv_signal_batch(surv_ch.begin() + l, surv_ch.begin() + (l + batch_size)); // Filtering will be performed on this
        array surv_signal_batch_ta(extended_surv_ch.begin() + l, extended_surv_ch.begin() + (l + batch_size + 2 * Na));

        mat x(batch_size, array(KD));
        mat x_ta(batch_size + 2 * Na, array(KD));

        // Generate matrices for filtering and estimation from the current batch
        int subspace_vector_index = -1;
        for (const vector<int>& k : subspace_list) {
            ++subspace_vector_index;

            array doppler_shift_array_ta, doppler_shift_array;
            for (int i = l - Na; i < l + batch_size + Na; ++i) {
                complex<double> to_append = static_cast<double>(i) * complex<double>(0, 1) * 2.0 * M_PI;
                to_append *= static_cast<double>(k[0]) / (2.0 * static_cast<double>(N));
                doppler_shift_array_ta.push_back(exp(to_append));
            }

            for (int i = l; l < l + batch_size; ++l) {
                complex<double> to_append = static_cast<double>(i) * complex<double>(0, 1) * 2.0 * M_PI;
                to_append *= static_cast<double>(k[0]) / (2.0 * static_cast<double>(N));
                doppler_shift_array_ta.push_back(exp(to_append));
            }

            if (!l) {
                assert(k[1] >= 0);
                array shifted_x = shift({ref_ch.begin() + l, ref_ch.begin() + l + batch_size}, k[1]);
                array shifted_x_ta = shift({extended_ref_ch.begin() + l, extended_ref_ch.begin() + (l + batch_size + 2 * Na)}, k[1]);

                for (int i = 0; i < batch_size; ++i) {
                    x[i][subspace_vector_index] = shifted_x[i];
                }

                for (int i = 0; i < 2 * Na + batch_size; ++i) {
                    x_ta[i][subspace_vector_index] = shifted_x_ta[i];
                }
            }
            else {
                for (int i = 0; i < batch_size; ++i) {
                    x[i][subspace_vector_index] = ref_ch[l - k[1] + i];
                }

                for (int i = 0; i < 2 * Na + batch_size; ++i) {
                    x_ta[i][subspace_vector_index] = extended_ref_ch[l - k[1] + i];
                }
            }

            for (int i = 0; i < batch_size; ++i) {
                x[i][subspace_vector_index] *= doppler_shift_array[i];
            }

            for (int i = 0; i < 2 * Na + batch_size; ++i) {
                x_ta[i][subspace_vector_index] *= doppler_shift_array_ta[i];
            }
        }

        mat surv_signal = transpose({surv_signal_batch});
        mat surv_signal_ta = transpose({surv_signal_batch_ta});
        assert(surv_signal.size() == 1 && surv_signal_ta.size() == 1);

        mat x_ta_H = transposeConj(x_ta);
        mat corr_matrix = matMult(x_ta_H, x_ta);
        mat corr_vector = matMult(x_ta_H, surv_signal_ta);

        mat weight_vector = matMult(inverse(corr_matrix), corr_vector);

        weight_vector = matMult(x, weight_vector);
        assert(weight_vector.size() == surv_signal.size());
        assert(weight_vector[0].size() == 1);
        int surv_signal_size = static_cast<int>(surv_signal.size());
        for (int i = 0; i < surv_signal_size; ++i) {
            filtered_surv_ch[l + i] = surv_signal[i][0] - weight_vector[i][0];
        }
    }

    return filtered_surv_ch;
}


array pruned_correlation(array& y, array& x, int clen) {

    int N = static_cast<int>(x.size());
    if (N != static_cast<int>(y.size()) || clen > (N + 1) / 2) {
        cerr << "ERROR: Invalid sized arrays. Exitting.\n";
        cerr << "N: " << N << "\n";
        cerr << "y size: " << y.size() << "\n";
        cerr << "clen: " << clen << "\n";
        cerr << "(N + 1) / 2" << (N + 1) / 2 << "\n";
        exit(1);
    }

    int cols = clen - 1;
    int rows = N / cols + 1;
    int zeropad = cols * rows - N;

    array x_with_padding = x;
    array y_with_padding = y;
    for (int i = 0; i < zeropad; ++i) {
        x_with_padding.push_back(0);
        y_with_padding.push_back(0);
    }

    // Set up input matrix, pad 0's if not multiple of correlation length
    mat xp(rows);
    mat yp(rows);

    int cnt = 0;
    for (int i = 0; i < N; i += cols) {
        xp[cnt] = array(x_with_padding.begin() + i, x_with_padding.begin() + (i + cols));
        yp[cnt++] = array(y_with_padding.begin() + i, y_with_padding.begin() + (i + cols));
    }
    assert(cnt == rows);
    
    // Pad matrices for FFT
    array zeros(cols);

    for (int i = 0; i < rows - 1; ++i) {
        yp[i].insert(yp[i].end(), yp[i + 1].begin(), yp[i + 1].end());
    }
    yp.back().insert(yp.back().end(), zeros.begin(), zeros.end());
    for (int i = 0; i < rows; ++i) {
        xp[i].insert(xp[i].end(), zeros.begin(), zeros.end());
    }

    // execute FFT on matrices
    for (int i = 0; i < rows; ++i) {
        fft(xp[i]);
        fft(yp[i]);
    }

    mat& xpw = xp;
    mat& bpw = yp;

    // magic formula
    mat bpw_conj = conjugate(bpw);
    mat mult_xpw_bpw = multiply(xpw, bpw_conj);
    for (int i = 0; i < rows; ++i) {
        ifft(mult_xpw_bpw[i]);
    }

    mat fft_shifted_conj = conjugate(fftshift(mult_xpw_bpw));

    // Sum cols and reverse
    array output(clen);
    for (unsigned int i = 0; i < fft_shifted_conj.size(); ++i) {
        int idx = 0;
        for (int j = clen - 1; j >= 0; --j) {
            output[idx++] += fft_shifted_conj[i][j];
        }
    }

    return output;
}


array Wiener_SMI_MRE(array& ref_ch, array& surv_ch, int K) {
    if (ref_ch.size() != surv_ch.size()) {
        cerr << "Mismatching channel sized\n Exitting.\n";
        exit(1);
    }

    int N = static_cast<int>(ref_ch.size());
    mat R(K, array(K));

    array pruned = pruned_correlation(ref_ch, ref_ch, K); // First column of autocorrelation matrix

    for (int i = 0; i < K; ++i) {
        R[i][0] = pruned[i];
    }

    array r = pruned_correlation(surv_ch, ref_ch, K); // Cross-correlation vector

    for (int k = 1; k < K; ++k) {
        array shifted = shift(pruned, k);
        for (int i = 0; i < K; ++i) {
            R[i][k] = shifted[i];
        }
    }

    R = add(R, transposeConj(R));
    for (int i = 0; i < K; ++i) {
        R[i][i] *= 0.5;
    }

    array w = dot(inverse(R), r);

    array output = surv_ch;
    array convolved = convolve(ref_ch, w);
    for (int i = 0; i < N; ++i) {
        // cerr << convolved[i] << "\n";
        output[i] -= convolved[i];
    }

    return output;
}