#include "detector.h"

using std::complex;
using std::copy;
using std::exp;
using std::valarray;
using std::vector;

// https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.signal.hann.html
vector<double> hann(int M) {
    vector<double> w(M);
    for (int n = 0; n < M; ++n) {
        w[n] = 0.5 - 0.5 * cos(double(2 * M_PI * n) / double(M - 1));
    }

    return w;
}

array windowing(const array& surv_ch, const string& window_function) {
    vector<double> windowed;
    if (window_function == "Hann") {
        windowed = hann(static_cast<int>(surv_ch.size()));
    }

    int M = static_cast<int>(surv_ch.size());
    array output(M);
    for (int i = 0; i < M; ++i) {
        output[i] = surv_ch[i] * windowed[i];
    }

    return output;
}

mat cc_detector_td(
    const array& ref_ch,
    const array& surv_ch,
    int fs,
    int fD_max,
    int fD_s,
    int r_max ) {
    
    // Doppler shifted copy of reference signal
    array ref_doppler_shift(ref_ch.size(), complex<double>(0.0, 0.0));
    vector<int> doppler_freqs;
    arange(doppler_freqs, -fD_max, fD_max + fD_s, fD_s);

    // Range-doppler matrix
    mat rd_matrix(doppler_freqs.size(), array(r_max, complex<double>(0.0, 0.0)));

    for (int fD_idx = 0; fD_idx < static_cast<int>(doppler_freqs.size()); ++fD_idx) {
        // Doppler Shift
        int ref_size = static_cast<int>(ref_ch.size());
        for (int i = 0; i < ref_size; ++i) {
            ref_doppler_shift[i] = ref_ch[i] * exp(static_cast<double>(i) * complex<double>(0, 1) * 2.0 * M_PI * static_cast<double>(doppler_freqs[fD_idx]) / static_cast<double>(fs));
        }

        // Ranged Loop
        for (int tau = 0; tau < r_max; ++tau) {
            rd_matrix[fD_idx][tau] = dot(conjugate(shift(ref_doppler_shift, tau)), surv_ch);
        }
    }

    return rd_matrix;
}

/*
 * Description: Overlap and Save method of cross correlation
 * 
 */
mat cc_detector_ons(
    array& ref_ch,
    array& surv_ch,
    int fs,
    int fD_max,
    int r_max) {

    // input check
    int N = ref_ch.size();
    assert(N % r_max == 0);

    // Set processing parameters
    double fD_step = static_cast<double>(fs) / static_cast<double>(2 * N);
    int doppler_freq_size = static_cast<int>(static_cast<double>(fD_max) /  fD_step);
    int no_sub_tasks = N / r_max;

    mat mx(2 * doppler_freq_size + 1, array(r_max, complex<double>(0.0, 0.0)));

    // Define ref_ch_align and surv_ch_align
    mat ref_ch_align = reshape(ref_ch, no_sub_tasks, r_max);
    mat surv_ch_align = reshape(surv_ch, no_sub_tasks, r_max);

    for (array& row : ref_ch_align) {
        row.resize(row.size() * 2);
    }

    for (int i = 0; i < no_sub_tasks - 1; ++i) {
        int row_size = surv_ch_align[i].size();
        surv_ch_align[i].resize(row_size * 2);
        for (int j = row_size; j < row_size * 2; ++j) {
            surv_ch_align[i][j] = surv_ch_align[i + 1][j - row_size];
        }
    }
    surv_ch_align.back().resize(surv_ch_align.back().size() * 2, complex<double>(0.0, 0.0));

    // vector<valarray<complex<double>>> ref_fft(ref_ch_align.size());
    // vector<valarray<complex<double>>> surv_fft(surv_ch_align.size());
    
    // Row wise fft on both channels
    for (array& row : ref_ch_align) {
        fft(row);
    }

    for (array& row : surv_ch_align) {
        fft(row);
    }

    assert(ref_ch_align.size() == surv_ch_align.size());

    // Correlation in the frequency domain
    mat corr = multiply(surv_ch_align, conjugate(ref_ch_align));

    //Renaming to stay consistent with python implementation
    for (array& row : corr) {
        ifft(row);
    }

    // Transpose the corr and append 0's to pad
    mat corr_transponse(2 * r_max, array(2 * no_sub_tasks));
    
    assert(2 * r_max == static_cast<int>(corr[0].size()) && no_sub_tasks == static_cast<int>(corr.size()));
    for (int i = 0; i < no_sub_tasks; ++i) {
        for (int j = 0; j < 2 * r_max; ++j) {
            corr_transponse[j][i] = corr[i][j];
        }
    }

    for (array& col : corr_transponse) {
        fft(col);
    }

    for (int i = 0; i < doppler_freq_size; ++i) {
        for (int j = 0; j < r_max; ++j) {
            mx[i][j] = corr_transponse[j][2 * no_sub_tasks - doppler_freq_size + i];
            mx[doppler_freq_size + i][j] = corr_transponse[j][i];
        }
    }
    
    for (int j = 0; j < r_max; ++j) {
        mx[doppler_freq_size][j] = corr_transponse[j][doppler_freq_size];
    }

    return mx;
}