#include "directionEstimation.h"
#include "numpy_helper.h"
#include <vector>
#include <complex>       
#include <math.h>
#include <algorithm>

#include <iostream>
#include <fstream>
#include <Eigen/Eigenvalues> 

typedef vector<complex<double>> my_arr;

using namespace std;

bool myfunction (pair<complex<double>,my_arr>& i,pair<complex<double>,my_arr>& j) { 
    return i.first.real() + i.first.imag() < j.first.real() + j.first.imag();
}


mat gen_ula_scanning_vectors(vector<double>& array_alignment, vector<double>& thetas) {
    int thetas_size = thetas.size();
    int M = array_alignment.size();
    const std::complex< double> i(0, 2*M_PI);
    mat scanning_vectors (M, my_arr(thetas.size(),0));
    for (int i2 = 0; i2 < thetas_size; i2++) {
        for (int j = 0; j < M; j++) {

            scanning_vectors[j][i2] = exp(i*array_alignment[j]*cos(0.0174533*thetas[i2]));
        }
    }
        
    return scanning_vectors;
}

my_arr DOA_MUSIC(mat& R, mat& scanning_vectors, int signal_dimension, int angle_resolution) {
    /*                                 
                    MUSIC - Multiple Signal Classification method
        
        
        Description:
         ------------    
           The function implements the MUSIC method for direction estimation
           
           Calculation method : 
                                                    1
                        ADORT(theta) = ---------------------------
                                             H        H 
                                      S(theta) * En En  * S(theta)
         Parameters:
        -----------                
            :param R: spatial correlation matrix            
            :param scanning_vectors : Generated using the my_arr alignment and the incident angles                                                 
            :param signal_dimension:  Number of signal sources    
                       
            :type R: 2D numpy my_arr with size of M x M, where M is the number of antennas in the antenna system                            
            :tpye scanning vectors: 2D numpy my_arr with size: M x P, where P is the number of incident angles                                   
            :type signal_dimension: int
            
       Return values:
       --------------
       
            :return  ADORT : Angular dependent orthogonality. Expresses the orthongonality of the current steering vector to the 
                    noise subspace
            :rtype : numpy my_arr
         
            :return -1, -1: Input spatial correlation matrix is not quadratic
            :return -2, -2: dimension of R not equal with dimension of the antenna my_arr 
            :return -3, -3: Spatial correlation matrix is singular
    */
    if (R.size()==0 || R.size() != R[0].size()) {
        cout << "ERROR: Correlation matrix is not quadratic" << endl;
        exit(1);
    }
    
    if (R.size() != scanning_vectors.size()) {
        cout << "ERROR: Correlation matrix dimension does not match with the antenna my_arr dimension" << endl;
        exit(1);
    }

    int R_size = R.size();
    Eigen::MatrixXcd mat_R(R.size(), R.size());
    for (int i = 0; i < R_size; i++){
        for (int j = 0; j < R_size; j++){
             mat_R.row(i)[j] = R[i][j];
        }
    }

    my_arr ADORT(scanning_vectors[0].size(),0);
    int M = R.size();
    
    // --- Calculation ---
    // Determine eigenvectors and eigenvalues

    Eigen::SelfAdjointEigenSolver<Eigen::MatrixXcd> ces;
    ces.compute(mat_R);
    Eigen::VectorXcd vals = ces.eigenvalues();
    Eigen::MatrixXcd vecs = ces.eigenvectors();

    vector<pair<complex<double>,my_arr> > eig_my_arr(M);
    for (int i = 0; i<M;i++) {
        eig_my_arr[i].first = abs(vals[i]);

        eig_my_arr[i].second = my_arr(vecs.rows(),0);
        for (int j = 0; j<vecs.rows(); j++) {
            eig_my_arr[i].second[j] = vecs.row(j)[i];
        }

    }
    sort (eig_my_arr.begin(), eig_my_arr.end(), myfunction); // 12 32 45 71(26 33 53 80)

    // Generate noise subspace matrix
    int noise_dimension = M - signal_dimension    ;

    mat E(M,my_arr(noise_dimension,0));
    int my_arr_size = eig_my_arr[0].second.size();

    for (int i=0; i < noise_dimension; i++) {
        for (int j = 0; j < my_arr_size; j++) {
            E[j][i] = eig_my_arr[i].second[j];
        }
    }

    int theta_index=0;
    int scanning_vector_rows = scanning_vectors.size();
    int scanning_vector_cols = scanning_vectors[0].size();

    for (int i = 0; i < scanning_vector_cols; i++) {
        mat S_theta_(scanning_vectors.size(),my_arr(1,0));
        for (int j = 0; j < scanning_vector_rows;j++) {
            S_theta_[j][0] = scanning_vectors[j][i];
        }

        ADORT[theta_index] = 1/abs( matMult(transposeConj(S_theta_) ,matMult(matMult(E, transposeConj(E)), S_theta_))[0][0]);
        theta_index+=1;
    }
     
    return ADORT;
 
}


vector<double> target_DOA_estimation(vector<mat>& rd_maps, vector<pair<int,int>>& hit_list, vector<double>& array_alignment) {
    vector<double> doa_list;
    vector<double> thetas(180);
    for(int i = 0; i < 180; i++) {
        thetas[i] = i;
    }
    mat scanning_vectors = gen_ula_scanning_vectors(array_alignment, thetas);
    int hit_list_size = hit_list.size();
    for (int i = 0; i < hit_list_size; i++) {
        my_arr azimuth_vector(rd_maps.size());
        int azimuth_vector_size = azimuth_vector.size();
        for (int j = 0; j < azimuth_vector_size; j++) {
            azimuth_vector[j] = rd_maps[j][hit_list[i].second][hit_list[i].first];
        }
        my_arr azi_conj = conjugate(azimuth_vector);
        mat R(azi_conj.size(),my_arr(azi_conj.size()));
        int azi_conj_size = azi_conj.size();
        for (int j = 0; j < azi_conj_size; j++) {
            for (int k = 0; k <  azi_conj_size; k++) {
                R[j][k] = azimuth_vector[j]*azi_conj[k];
            }
        }
        my_arr doa_res = DOA_MUSIC(R, scanning_vectors, 1);
        double max = 0;
        int argmax = 0;
        int doa_res_size = doa_res.size();
        for (int j = 0; j < doa_res_size; j++) {
            if (doa_res[j].real() + doa_res[j].imag() > max) {
                max = doa_res[j].real() + doa_res[j].imag();
                argmax = j;
            }
        }
        doa_list.push_back(argmax);
    }
    return doa_list;
}