#include <algorithm>
#include <chrono>
#include <cmath>
#include <complex>
#include <fstream>
#include <iostream>
#include <limits>
#include <string>
#include <vector>

#include "beamform.h"
#include "clutterCancellation.h"
#include "channelPreparation.h"
#include "detector.h"
#include "directionEstimation.h"
#include "helper.h"
#include "metricExtract.h"
#include "numpy_helper.h"

using std::ceil;
using std::cerr;
using std::complex;
using std::cout;
using std::ifstream;
using std::log2;
using std::max;
using std::max_element;
using std::numeric_limits;
using std::pow;
using std::round;
using std::string;
using std::to_string;
using std::vector;

using namespace std::chrono;

typedef vector<int> int_array;
typedef vector<double> double_array;
typedef vector<double_array> double_mat;
typedef vector<complex<double>> array;
typedef vector<array> mat;
typedef vector<mat> tens;

#define NANOSECONDS_IN_SECOND 1000000000.0

/*
 * Description:
 * Target reference track stores relevent information for each dataset entry.
 * 
 * Data was ported from this dataset: https://www.tamaspeto.com/Vega/VEGAM20191219K4U0C0S4DVBT
 * 
 * Format:
 * 
 * <num_rows> <num_cols>
 * <Time index> <Timestamp> <Latitude> <Longitude> <Altitude> <Speed> <Direction> <Range> <Doppler> <Azimuth>
 * 
 * The first two lines determine the number of time indices and the number of elements in a row (should always be 10)
 */
void load_target_track(double_mat& target_reference_track) {
    ifstream infile("dataset/target_reference_track.txt");
    if (!infile.is_open()) {
        cerr << "ERROR: Cannot open target reference track file.\n";
        exit(1);
    }

    int rows, cols;
    infile >> rows >> cols;

    target_reference_track = double_mat(rows, double_array(cols));
    for (int i = 0; i < rows; ++i) {
        for (int j = 0; j < cols; ++j) {
            infile >> target_reference_track[i][j];
        }
    }
}

/*
 * Description:
 * Loads a single time index into a matrix.
 * 
 * Data was ported from this dataset: https://www.tamaspeto.com/Vega/VEGAM20191219K4U0C0S4DVBT
 * 
 * Format:
 * <Sampling Frequency>
 * <num_rows> <num_cols>
 * real_1 imag_1 real_2 imag_2 ... real_num_cols imag_num_cols
 * ...
 * 
 * The first line denotes the sampling frequency.
 * The second line denotes the number of rows and columns in the matrix.
 * The number of rows denotes the number of antennas we have. We assume the reference channel is at index 1.
 * Each row is then the data collected by that antenna.
 */
void load_dataset(const string& filename, mat& data, double& fs) {
    ifstream infile(filename);

    if (!infile.is_open()) {
        cerr << "ERROR: Cannot open dataset file.\n";
        exit(1);
    }

    infile >> fs;

    int rows, cols;
    infile >> rows >> cols;

    data = mat(rows, array(cols));
    for (int i = 0; i < rows; ++i) {
        for (int j = 0; j < cols; ++j) {
            double real, imag;
            infile >> real >> imag;
            data[i][j] = complex<double>(real, imag);
        }
    }
}

int main() {
    int ref_ch_index = 0;
    int K = 128; // Time domain filter dimension
    int M = 3; // Number of antenna channels
    string cc_det_window = "Hann";
    double d = 0.528; // Antenna inter-element spacing

    // Detection performanace metrics extraction parameters
    int_array win = {6, 6, 3, 3};
    int_array win_pos = {30, 50};
    int search_window_size = 8; // One sided

    double_mat target_reference_track;
    load_target_track(target_reference_track);

    int start_file_idx = int(target_reference_track.front()[0]);
    int stop_file_idx = int(target_reference_track.back()[0]);
    int total_indices = stop_file_idx - start_file_idx + 1; //+1 since both endpoints are included

    double max_range = -numeric_limits<double>::infinity();
    double max_Doppler = max_range;
    for (double_array& row : target_reference_track) {
        max_range = max(max_range, row[7]);
        max_Doppler = max(max_Doppler, row[8]);
    }

    double_array array_alignment(M);
    for (int i = 0; i < M; ++i) {
        array_alignment[i] = i * d;
    }

    tens rd_maps_for_doa;

    double isolation_time, beamform_time, td_filter_time, detector_time, target_find_time, doa_est_time, iteration_time;
    
    double isolation_time_total, beamform_time_total, td_filter_time_total, detector_time_total, target_find_time_total, doa_est_time_total, iteration_time_total;
    isolation_time_total = 0;
    beamform_time_total = 0;
    td_filter_time_total = 0;
    detector_time_total = 0;
    target_find_time_total = 0;
    doa_est_time_total = 0;
    iteration_time_total = 0;
    
    double isolation_time_max, beamform_time_max, td_filter_time_max, detector_time_max, target_find_time_max, doa_est_time_max, iteration_time_max;
    isolation_time_max = 0;
    beamform_time_max = 0;
    td_filter_time_max = 0;
    detector_time_max = 0;
    target_find_time_max = 0;
    doa_est_time_max = 0;
    iteration_time_max = 0;

    for (int t = 0; t < stop_file_idx - start_file_idx + 1; ++t) {
        cout << "====================== Processing Time Index " << t + start_file_idx << " ======================\n\n";

        auto iter_start = high_resolution_clock::now();
        // Load information
        mat dataset;
        double fs;
        load_dataset("dataset/dataset_" + to_string(t + start_file_idx) + ".txt", dataset, fs);

        // CPI size
        int N = static_cast<int>(dataset[0].size());

        // Doppler resolution
        double fD_res = static_cast<double>(fs) / static_cast<double>(2 * N);
        double range_res = (3.0 * 100000000.0) / fs; 
        int Doppler_cells = int(max_Doppler / fD_res) + search_window_size + win[0] + 1;
        int max_Doppler_ext = ceil(Doppler_cells * fD_res);
        int max_range_ext = int(pow(2, ceil(log2((max_range / range_res) + search_window_size + win[0]))));

        int ref_target_range_cell = int(round(target_reference_track[t][7] / range_res));
        int ref_target_Doppler_cell = int(round(target_reference_track[t][8] / fD_res) + Doppler_cells);
        double ref_target_doa_dir = target_reference_track[t][9];

        // Uniform linear array assumption
        if (ref_target_doa_dir < 0)
            ref_target_doa_dir += 180;
        if (ref_target_doa_dir > 180)
            ref_target_doa_dir -= 180;

        auto start = high_resolution_clock::now();
        pair<array, mat> ref_and_survs = isolate_channels(dataset, ref_ch_index);
        isolation_time = (high_resolution_clock::now() - start).count() / NANOSECONDS_IN_SECOND;
        isolation_time_total += isolation_time;
        if (isolation_time > isolation_time_max) isolation_time_max = isolation_time;
        array& ref_ch = ref_and_survs.first;
        mat& surv_chs = ref_and_survs.second;
        mat surv_chs_t = transpose(surv_chs);

        start = high_resolution_clock::now();
        array surv_ch = beamform_surveillance(surv_chs_t, array_alignment, ref_target_doa_dir);
        beamform_time = (high_resolution_clock::now() - start).count() / NANOSECONDS_IN_SECOND;
        beamform_time_total += beamform_time;
        if (beamform_time > beamform_time_max) beamform_time_max = beamform_time;

        start = high_resolution_clock::now();
        array filt_surv_ch = time_domain_filter_surveillance(ref_ch, surv_ch, K);
        td_filter_time = (high_resolution_clock::now() - start).count() / NANOSECONDS_IN_SECOND;
        td_filter_time_total += td_filter_time;
        if (td_filter_time > td_filter_time_max) td_filter_time_max = td_filter_time;

        start = high_resolution_clock::now();
        array windowed_surv_ch = windowing(filt_surv_ch, cc_det_window);
        mat rd_matrix = cc_detector_ons(ref_ch, windowed_surv_ch, fs, max_Doppler_ext, max_range_ext);
        detector_time = (high_resolution_clock::now() - start).count() / NANOSECONDS_IN_SECOND;
        detector_time_total += detector_time;
        if (detector_time > detector_time_max) detector_time_max = detector_time;

        start = high_resolution_clock::now();
        pair<int, int> results = find_target(rd_matrix,
                    ref_target_range_cell - search_window_size,
                    ref_target_range_cell + search_window_size,
                    ref_target_Doppler_cell - search_window_size,
                    ref_target_Doppler_cell + search_window_size);
        target_find_time = (high_resolution_clock::now() - start).count() / NANOSECONDS_IN_SECOND;
        target_find_time_total += target_find_time;
        if (target_find_time > target_find_time_max) target_find_time_max = target_find_time;
        
        int target_range_cell = results.first;
        int target_doppler_cell = results.second;

        vector<pair<int,int>> hit_list = {{target_range_cell, target_doppler_cell}};

        start = high_resolution_clock::now();
        if (t == 0) {
            rd_maps_for_doa = tens(surv_chs.size());
        }

        for (unsigned int m = 0; m < surv_chs.size(); ++m) {
            filt_surv_ch = time_domain_filter_surveillance(ref_ch, surv_chs[m], K);
            windowed_surv_ch = windowing(filt_surv_ch, cc_det_window);
            rd_maps_for_doa[m] = cc_detector_ons(ref_ch, windowed_surv_ch, fs, max_Doppler_ext, max_range_ext);
        }

        double_array estimated_doas = target_DOA_estimation(rd_maps_for_doa, hit_list, array_alignment);

        doa_est_time = (high_resolution_clock::now() - start).count() / NANOSECONDS_IN_SECOND;
        doa_est_time_total += doa_est_time;
        if (doa_est_time > doa_est_time_max) doa_est_time_max = doa_est_time;

        cout << "Target range cell ref/found [" << ref_target_range_cell << "/" << target_range_cell << "]\n";
        cout << "Target Doppler cell ref/found [" << ref_target_Doppler_cell << "/" << target_doppler_cell << "]\n";
        cout << "Target azimuth ref/found [" << target_reference_track[t][9] << "/" << estimated_doas[0] << "]\n";

        int_array target_rd = {target_range_cell, target_doppler_cell};

        target_rd[1] -= (static_cast<int>(rd_matrix.size()) - 1) / 2;
        win_pos[1] -= (static_cast<int>(rd_matrix.size()) - 1) / 2;

        double delta = extract_delta(rd_matrix, target_rd, win);
        cout << "Target SINR " << delta << "\n";

        iteration_time = (high_resolution_clock::now() - iter_start).count() / NANOSECONDS_IN_SECOND;
        iteration_time_total += iteration_time;
        if (iteration_time > iteration_time_max) iteration_time_max = iteration_time;

        cout << "\n";
        cout << "Time to isolate channels: " << isolation_time << " seconds\n";
        cout << "Time for beamforming: " << beamform_time << " seconds\n";
        cout << "Time to perform time-domain filter surveillance : " << td_filter_time << " seconds\n";
        cout << "Time to detect: " << detector_time << " seconds\n";
        cout << "Time to find target: " << target_find_time << " seconds\n";
        cout << "Time to estimate direction of arrival: " << doa_est_time << " seconds\n";
        cout << "Total time to process time index " << t + start_file_idx << ": " << iteration_time  << " seconds\n";
        cout << "\n";

    }

    cout << "====================== RESULTS SUMMARY ======================\n\n";
    cout << "Avg time to isolate channels: " << isolation_time_total/total_indices << " seconds\n";
    cout << "Avg time for beamforming: " << beamform_time_total/total_indices << " seconds\n";
    cout << "Avg time to perform time-domain filter surveillance : " << td_filter_time_total/total_indices << " seconds\n";
    cout << "Avg time to detect: " << detector_time_total/total_indices << " seconds\n";
    cout << "Avg time to find target: " << target_find_time_total/total_indices << " seconds\n";
    cout << "Avg time to estimate direction of arrival: " << doa_est_time_total/total_indices << " seconds\n";
    cout << "Avg total time to process single time index: " << iteration_time_total/total_indices  << " seconds\n";
    cout << "\n";
    cout << "Max time to isolate channels: " << isolation_time_max << " seconds\n";
    cout << "Max time for beamforming: " << beamform_time_max << " seconds\n";
    cout << "Max time to perform time-domain filter surveillance : " << td_filter_time_max << " seconds\n";
    cout << "Max time to detect: " << detector_time_max << " seconds\n";
    cout << "Max time to find target: " << target_find_time_max << " seconds\n";
    cout << "Max time to estimate direction of arrival: " << doa_est_time_max << " seconds\n";
    cout << "Max total time to process single time index: " << iteration_time_max  << " seconds\n";
    cout << "\n";

    return 0;
}
